from shimoku.api.resources.reports.tabs_group import TabsGroup
from shimoku.utils import create_function_name

from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen. \
        report_types_code_gen.code_gen_from_other import code_gen_from_other_reports, delete_default_properties
from shimoku.code_gen.utils_code_gen import code_gen_from_value
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.reports_code_gen.code_gen_from_reports import AppCodeGen


async def code_gen_tabs_and_other(
        self: 'AppCodeGen', tree: dict
) -> list[str]:
    """ Generate code for tabs and other components.
    :param tree: tree of reports
    :return: list of code lines
    """
    code_lines: list[str] = []
    reports_ordered = sorted(tree['other'] + tree['tab_groups'], key=lambda x: x['order'])
    for i, report in enumerate(reports_ordered):
        report_obj = report if not isinstance(report, dict) else report['tabs_group']
        if len(report_obj['bentobox']):
            if (self._actual_bentobox is None or
                    self._actual_bentobox['bentoboxId'] != report_obj['bentobox']['bentoboxId']):
                self._actual_bentobox = report_obj['bentobox']

                cols_size = self._actual_bentobox['bentoboxSizeColumns']
                rows_size = self._actual_bentobox['bentoboxSizeRows']
                code_lines.extend([
                    '',
                    f'shimoku_client.plt.set_bentobox(cols_size={cols_size}, rows_size={rows_size})'
                ])
        elif self._actual_bentobox is not None:
            self._actual_bentobox = None
            code_lines.append('shimoku_client.plt.pop_out_of_bentobox()')

        if isinstance(report, dict):
            code_lines.extend([
                '',
                f'tabs_group_{create_function_name(report["name"])}(shimoku_client)']
            )
        else:
            code_lines.extend(
                await code_gen_from_other_reports(self, report)
            )

        if i == len(reports_ordered) - 1 and self._actual_bentobox is not None:
            self._actual_bentobox = None
            code_lines.append('shimoku_client.plt.pop_out_of_bentobox()')

    return code_lines


async def code_gen_from_tabs_group(
        self: 'AppCodeGen', tree: dict, is_last: bool = False
) -> list[str]:
    """ Generate code for a tabs group.
    :param tree: tree of reports
    :param is_last: whether the tabs group is the last one
    :return: list of code lines
    """
    code_lines = []
    tabs_keys = list(tree['tabs'].keys())
    if len(tabs_keys) == 0:
        return ['pass']
    tabs_group: TabsGroup = tree['tabs_group']
    tabs_index = (tree['name'], tabs_keys[0])
    parent_tabs_index = tree['parent_tabs_index']
    properties = delete_default_properties(tabs_group['properties'], TabsGroup.default_properties)
    del properties['hash']
    if 'tabs' in properties:
        del properties['tabs']
    if 'variant' in properties:
        properties['just_labels'] = True
        del properties['variant']

    for tab in tree['tabs']:
        code_lines.extend(['', f'def tab_{create_function_name(tabs_index[0])}_{create_function_name(tab)}():'])
        # tab_code = await code_gen_tabs_functions(tree['tabs'][tab]['tab_groups'])
        tab_code = await code_gen_tabs_and_other(self, tree['tabs'][tab])
        if not tab_code:
            tab_code = ['pass']
        code_lines.extend([f'    {line}' for line in tab_code])

    code_lines.extend([
        '',
        'shimoku_client.plt.set_tabs_index(',
        f'    tabs_index=("{tabs_index[0]}", "{tabs_index[1]}"), order={tabs_group["order"]}, ',
    ])
    if tabs_group['sizeColumns']:
        code_lines.append(f'    cols_size={tabs_group["sizeColumns"]},')
    if tabs_group['sizeRows']:
        code_lines.append(f'    rows_size={tabs_group["sizeRows"]},')
    if tabs_group['sizePadding']:
        code_lines.append(f'    padding="{tabs_group["sizePadding"]}",')
    if parent_tabs_index:
        code_lines.extend([f'    parent_tabs_index={parent_tabs_index},'])
    code_lines.extend([f'    {k}={code_gen_from_value(v)},' for k, v in properties.items()])
    code_lines.extend([')'])

    for tab in tree['tabs']:
        code_lines.extend(['', f'shimoku_client.plt.change_current_tab("{tab}")'])
        code_lines.extend([f'tab_{create_function_name(tabs_index[0])}_{create_function_name(tab)}()'])

    if parent_tabs_index:
        if not is_last:
            code_lines.extend([
                '',
                f'shimoku_client.plt.set_tabs_index(("{parent_tabs_index[0]}", "{parent_tabs_index[1]}"))'
            ])
    else:
        code_lines.extend(['', 'shimoku_client.plt.pop_out_of_tabs_group()'])

    return code_lines


async def code_gen_tabs_functions(self, path: str) -> list[str]:
    """ Generate code for tabs groups functions.
    :return: list of code lines
    """
    code_lines = []
    if path not in self._code_gen_tree.all_tab_groups:
        return code_lines
    for tabs_group in self._code_gen_tree.all_tab_groups[path]:
        tab_code_lines = await code_gen_from_tabs_group(self, tabs_group)
        code_lines.extend([
            '',
            f'def tabs_group_{create_function_name(tabs_group["name"])}'
            f'(shimoku_client: Client):',
            *['    ' + line for line in tab_code_lines]
        ])
    return code_lines



import asyncio

from typing import Optional

from shimoku.api.resources.role import user_delete_role, user_get_role, user_get_roles, user_create_role
from shimoku.api.resources.business import Business
from shimoku.api.resources.app import App
from shimoku.api.resources.report import Report
from shimoku.api.resources.universe import Universe
from shimoku.utils import create_normalized_name
from shimoku.exceptions import WorkspaceError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging
logger = logging.getLogger(__name__)


class WorkspacesLayer(ClassWithLogging):
    """
    This class is used to interact with the API at the Workspace layer
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self, universe: Universe,
    ):
        self._universe = universe
        self._get_for_roles = self._get_business_with_warning

    async def _check_similar_name(self, name: str):
        """
        Check if there is a workspace with a similar name
        :param name: name of the workspace
        """
        all_business = await self._universe.get_businesses()
        normalized_name = create_normalized_name(name)
        if normalized_name in [create_normalized_name(business['name']) for business in all_business]:
            log_error(
                logger,
                'There exists a workspace with a very similar name, please choose another name',
                WorkspaceError
            )

    async def _get_business_with_warning(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[Business]:
        """
        Get a workspace with a warning if it does not exist
        :param name: name of the workspace
        :param uuid: uuid of the workspace
        """
        business: Business = await self._universe.get_business(uuid=uuid, name=name)
        if not business:
            logger.warning(f"Workspace ({name if name else uuid}) not found")
        return business

    async def create_workspace(
        self, name: str, create_default_roles: bool = True,
        theme: Optional[dict] = None
    ) -> dict:
        """Create a new workspace and the necessary roles if specified
        :param name: Name of the workspace
        :param create_default_roles: Create the default roles for the workspace
        :param theme: Theme of the workspace
        :return: workspace data
        """
        await self._check_similar_name(name)

        business = await self._universe.create_business(name=name, theme=theme)
        logger.info(f'Created workspace ({str(business)})')

        if create_default_roles:
            create_roles_tasks = []

            for role_permission_resource in ['DATA', 'DATA_EXECUTION', 'USER_MANAGEMENT', 'BUSINESS_INFO']:
                create_roles_tasks.append(
                    business.create_role(
                        role='business_read',
                        resource=role_permission_resource,
                    )
                )

            await asyncio.gather(*create_roles_tasks)
            logger.info(f'Created default roles for workspace ({str(business)})')

        return business.cascade_to_dict()

    async def update_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        new_name: Optional[str] = None, theme: Optional[dict] = None
    ) -> bool:
        """Update a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param new_name: New name of the workspace
        :param theme: New theme of the workspace
        :return: True if the workspace was updated
        """
        params = {}

        if new_name is not None:
            await self._check_similar_name(new_name)
            params['new_name'] = new_name

        if isinstance(theme, dict):
            params['theme'] = theme

        return await self._universe.update_business(uuid=uuid, name=name, **params)

    async def delete_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> bool:
        """Delete a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: True if the workspace was deleted
        """
        return await self._universe.delete_business(uuid=uuid, name=name)

    async def get_workspace(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> dict:
        """Get a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: Workspace data
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return {}
        return business.cascade_to_dict()

    async def get_workspace_menu_paths(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of apps
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [app.cascade_to_dict() for app in await business.get_apps()]

    async def get_workspace_menu_path_ids(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the menu path ids of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of app ids
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [app['id'] for app in await business.get_apps()]

    async def get_workspace_boards(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> list[dict]:
        """Get the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of dashboards
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return []
        return [dashboard.cascade_to_dict() for dashboard in await business.get_dashboards()]

    async def delete_all_workspace_menu_paths(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """Delete all menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        apps = await business.get_apps()
        await asyncio.gather(*[business.delete_app(uuid=app['id']) for app in apps])

    async def delete_all_workspace_boards(
        self, uuid: Optional[str] = None, name: Optional[str] = None,
        force: bool = False
    ):
        """Delete all boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param force: Flag to force delete the boards
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        dashboards = await business.get_dashboards()
        if force:
            await asyncio.gather(*[dashboard.remove_all_apps() for dashboard in dashboards])
        await asyncio.gather(*[business.delete_dashboard(uuid=dashboard['id']) for dashboard in dashboards])

    async def change_boards_order(
        self, boards: list[str], uuid: Optional[str] = None, name: Optional[str] = None,
    ):
        """Change the order of the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param boards: list of dashboard names
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        await asyncio.gather(*[business.update_dashboard(name=d_name, order=i+1) for i, d_name in enumerate(boards)])

    @staticmethod
    async def _change_sub_paths_order(
        business: Business, menu_path: str, sub_paths: list[str]
    ):
        """
        Change the order of the sub paths of a menu path
        :param business: Business object
        :param menu_path: Menu path name
        :param sub_paths: list of sub paths
        """
        app: App = await business.get_app(name=menu_path, create_if_not_exists=False)
        if not app:
            logger.warning(f"Menu path ({menu_path}) not found")
            return
        reports: list[Report] = await app.get_reports()
        sub_paths = [create_normalized_name(sub_path) for sub_path in sub_paths]
        non_referenced_reports = sorted(
            [report for report in reports if report['path'] is not None and
             create_normalized_name(report['path']) not in sub_paths],
            key=lambda x: x['pathOrder']
        )
        for report in non_referenced_reports:
            sub_path = report['path']
            if sub_path not in sub_paths:
                sub_paths.append(sub_path)

        tasks = []
        for i, sub_path in enumerate(sub_paths):
            _reports = [report for report in reports if (report['path'] is not None and
                        create_normalized_name(report['path']) == create_normalized_name(sub_path))]
            for report in _reports:
                tasks.append(app.update_report(uuid=report['id'], pathOrder=i))

        logger.info(f'Updating {len(reports)} components from menu path ({str(app)})')

        await asyncio.gather(*tasks)

    async def change_menu_order(
        self, menu_order: list, uuid: Optional[str] = None, name: Optional[str] = None
    ):
        """Change the order of the menu of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param menu_order: list of menu names
        """
        business = await self._get_business_with_warning(uuid=uuid, name=name)
        if not business:
            return
        tasks = []
        for i, menu_option in enumerate(menu_order):
            menu_name = menu_option
            if isinstance(menu_option, tuple):
                menu_name, sub_paths = menu_option
                tasks.append(
                    self._change_sub_paths_order(
                        business=business,
                        menu_path=menu_name,
                        sub_paths=sub_paths
                    )
                )
            tasks.append(business.update_app(name=menu_name, order=i))

        await asyncio.gather(*tasks)

    # Role management
    get_role = user_get_role
    get_roles = user_get_roles
    create_role = user_create_role
    delete_role = user_delete_role



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class WorkspacesLayerHeader:
    """
    This class is used to interact with the API at the Workspace layer
    """

    def change_boards_order(
        self,
        boards: list,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Change the order of the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param boards: list of dashboard names
        """
        pass

    def change_menu_order(
        self,
        menu_order: list,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Change the order of the menu of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param menu_order: list of menu names
        """
        pass

    def create_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        role_name: Optional[str] = None,
        permission: Optional[str] = None,
        target: Optional[str] = None,
    ) -> dict:
        """
    Create a new role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param resource: resource of the role
    :param role_name: name of the role
    :param permission: permission of the role
    :param target: target of the role
    """
        pass

    def create_workspace(
        self,
        name: str,
        create_default_roles: bool = True,
        theme: Optional[dict] = None,
    ) -> dict:
        """Create a new workspace and the necessary roles if specified
        :param name: Name of the workspace
        :param create_default_roles: Create the default roles for the workspace
        :param theme: Theme of the workspace
        :return: workspace data
        """
        pass

    def delete_all_workspace_boards(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        force: bool = False,
    ):
        """Delete all boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param force: Flag to force delete the boards
        """
        pass

    def delete_all_workspace_menu_paths(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Delete all menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        """
        pass

    def delete_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_id: Optional[str] = None,
        role_name: Optional[str] = None,
    ):
        """
    Delete the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_id: id of the role
    :param role_name: name of the role
    """
        pass

    def delete_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """Delete a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: True if the workspace was deleted
        """
        pass

    def get_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_name: Optional[str] = None,
    ) -> list:
        """
    Get the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_name: name of the role
    """
        pass

    def get_roles(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
    Get the roles at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    """
        pass

    def get_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """Get a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: Workspace data
        """
        pass

    def get_workspace_boards(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the boards of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of dashboards
        """
        pass

    def get_workspace_menu_path_ids(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the menu path ids of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of app ids
        """
        pass

    def get_workspace_menu_paths(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """Get the menu paths of a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :return: list of apps
        """
        pass

    def update_workspace(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        theme: Optional[dict] = None,
    ) -> bool:
        """Update a workspace
        :param name: Name of the workspace
        :param uuid: UUID of the workspace
        :param new_name: New name of the workspace
        :param theme: New theme of the workspace
        :return: True if the workspace was updated
        """
        pass



import inspect
import importlib
import os
from shimoku.utils import get_args_with_defaults
base_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
target_paths = {
    os.path.join(base_path, 'src', 'shimoku', 'api', 'user_access_classes'): '*',
    os.path.join(base_path, 'src', 'shimoku', 'ai'): 'ai_layer.py',
    os.path.join(base_path, 'src', 'shimoku', 'plt'): 'plt_layer.py',
}


def handle_annotation(annotation: type) -> str:
    if hasattr(annotation, '_use_info_logging') and annotation._use_info_logging:
        return f'{annotation.__name__}Header'
    if annotation is type(None) or annotation is None:
        type_str = 'None'
    elif hasattr(annotation, '__name__'):
        type_str = annotation.__name__
    else:
        type_str = str(annotation)
    return (type_str.replace('NoneType', 'None')
                    .replace('pandas.core.frame.DataFrame', ' DataFrame')
                    .replace('typing.', '')
                    .replace('<built-in function callable>', 'callable')
            )


def handle_parameter(name: str, annotation: type, default_args: dict) -> str:
    if name in default_args:
        default_value = default_args[name]
        if isinstance(default_value, str):
            default_value = f'"{default_value}"'
        return f'        {name}: {handle_annotation(annotation)} = {default_value},'
    else:
        return f'        {name}: {handle_annotation(annotation)},'


for target_path, target_file in target_paths.items():
    output_path = os.path.join(target_path, 'generated_headers')
    if not os.path.exists(output_path):
        os.makedirs(output_path)
    for file in os.listdir(target_path):
        if file != target_file and target_file != '*':
            continue
        if file.endswith('.py') and not file.startswith('__'):
            print(f'{os.path.relpath(target_path, base_path).replace(os.sep, ".")[4:]}.{file[:-3]}')
            module = importlib.import_module(
                f'{os.path.relpath(target_path, base_path).replace(os.sep, ".")[4:]}.{file[:-3]}'
            )
            classes_for_generation = []
            for name, obj in inspect.getmembers(module):
                if not inspect.isclass(obj):
                    continue
                if name.startswith('_'):
                    continue
                if not hasattr(obj, '_use_info_logging') or not obj._use_info_logging:
                    continue
                classes_for_generation.append((name, obj))

            if not classes_for_generation:
                continue

            classes_code_lines = []
            for class_for_generation in classes_for_generation:
                print(class_for_generation[0])
                methods = inspect.getmembers(class_for_generation[1], predicate=inspect.isfunction)
                methods_code_lines = []
                for method in methods:
                    if method[0].startswith('_'):
                        continue
                    methods_code_lines.extend([
                        f'    def {method[0]}(',
                        f'        self,',
                        *[
                            handle_parameter(arg, method[1].__annotations__[arg], get_args_with_defaults(method[1]))
                            for arg in method[1].__annotations__ if arg not in ['self', 'return']
                        ]
                    ])
                    sig = inspect.signature(method[1])
                    params = sig.parameters.values()
                    has_kwargs = any([True for p in params if p.kind == p.VAR_KEYWORD])
                    if has_kwargs:
                        methods_code_lines.append(f'        **kwargs,')
                    if 'return' in method[1].__annotations__:
                        methods_code_lines.append(f'    ) -> {handle_annotation(method[1].__annotations__["return"])}:')
                    else:
                        methods_code_lines.append(f'    ):')
                    methods_code_lines.extend([
                        f'        """{method[1].__doc__}"""',
                        f'        pass',
                        f'',
                    ])

                class_code_lines = [
                    '',
                    f'class {class_for_generation[0]}Header:',
                    f'    """{class_for_generation[1].__doc__}"""',
                    '',
                    *methods_code_lines,
                ]
                classes_code_lines.extend(class_code_lines)
            main_code_lines = [
                '# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!',
                '# If the user access files are modified, this file has to be regenerated with the script.',
                'from typing import Optional, Union',
                'from pandas import DataFrame',
                f'',
                *classes_code_lines,
            ]
            with open(os.path.join(output_path, f'{classes_for_generation[-1][0]}Header.py'), 'w') as f:
                f.write('\n'.join(main_code_lines))
